""" MambuPy's Mambu REST API v2 package.

Objects using `Mambu REST API <https://api.mambu.com/?python>`_

.. autosummary::
   :toctree: _autosummary

    MambuPy.api.classes
    MambuPy.api.entities
    MambuPy.api.interfaces
    MambuPy.api.vos
    MambuPy.api.mambustruct
    MambuPy.api.mambuproduct
    MambuPy.api.mambuloan
    MambuPy.api.mambutransaction
    MambuPy.api.mambugroup
    MambuPy.api.mambuclient
    MambuPy.api.mambubranch
    MambuPy.api.mambucentre
    MambuPy.api.mambuuser
    MambuPy.api.mambutask
    MambuPy.api.mambucustomfield

Entities use the Mixin pattern to acquire functionality.
"""
